"""Private miscellaneous functions and classes. Do not use them !"""
import math
import datetime
from collections import namedtuple
from paradict import errors, const


# TODO: update this module (write/correct docstrings, etc), and write tests

IntParts = namedtuple("IntParts", ["sign", "prefix", "leading_zeros", "val"])
FloatParts = namedtuple("FloatParts", ["sign", "left_significand",
                                       "right_significand", "exponent"])


def make_indent_str(indents):
    """The type of 'indents' is int !"""
    return const.INDENT_STR * indents


def forge_bin(*items):
    """items are int, bytes, bytearrays, or Nones, returns a bytearray"""
    # items are integers
    cache = bytearray()
    for item in items:
        # Skip null value
        if item is None:
            continue
        # process item
        if isinstance(item, int):
            if item < 0:
                raise errors.Error("Negative integer isn't allowed")
            n = calc_uint_bytes(item)
            item = item.to_bytes(n, "little", signed=False)
        elif isinstance(item, (bytes, bytearray)):
            if not item:
                continue
        else:
            msg = "The forge_bin func only accepts ints, bytes and bytearrays as items"
            raise errors.Error(msg)
        cache.extend(item)
    return cache


def stringify_bin(b, offset=0, width=None, spaced=False):
    """Good for debug. Stringify some binary data"""
    cache = list()
    for x in b:
        h = hex(x)[2:]
        if len(h) != 2:
            h = "0" + h
        h = r"\x" + h
        cache.append(h)
    space = " " if spaced else ""
    width = width + offset if width else len(b)
    return space.join(cache[offset:width])


def prettify_grid(grid):
    """prettify grids, returns a list that can be iterated like this:
    for row in result:
        print("  ".join(row))"""
    r = list()
    for col in zip(*grid):
        cache = list()
        n = 0
        # stringifying
        for row in col:
            row = str(row)
            if len(row) > n:
                n = len(row)
            cache.append(row)
        # padding
        for i, row in enumerate(cache):
            if len(row) < n:
                row += (" "*(n-len(row)))
                cache[i] = row
        r.append(cache)
    return list(zip(*r))


def prettify_base16(s):
    """Prettify base16 string. Returns a list of strings, each string representing
        a line of 16 bytes"""
    return make_multiline(s, group_size=2, row_size=16)


def make_multiline(s, group_size=0, row_size=42):
    group_size = 0 if group_size <= 0 else group_size
    row_size = 42 if row_size <= 0 else row_size
    result = list()
    row = list()
    group = list()
    spacing = " " if group_size else ""
    for char in s:
        if group_size:
            group.append(char)
            if len(group) == group_size:
                row.append("".join(group))
                group = list()
        else:
            row.append(char)
        if len(row) == row_size:
            result.append(spacing.join(row))
            row = list()
    if row:
        result.append(spacing.join(row))
    return result


def strip_block_extra_space(block):
    """
    This function is useful for textual Paradict's text and raw.
    It strips extra space out from text/raw block
    which is either a list of strings or a string.
    Return a string.
    """
    three_dashes = "---"
    end_of_block = "\n" + three_dashes
    if not isinstance(block, str):
        block = "\n".join(block)
    if block == three_dashes:
        return ""
    block = block.rstrip()
    x = len(end_of_block)
    if block.endswith(end_of_block):
        return block[0:-x]
    return block


def calc_uint_bytes(x):
    """Return the number of bytes needed for a given integer"""
    return math.ceil(x.bit_length() / 8.0)


def deconstruct_datetime(dt):
    years = dt.year - const.COVID_YEAR
    tz = None
    if dt.tzinfo is not None:
        tz = dt.tzinfo
    if years >= 0:
        xdt = datetime.datetime(year=dt.year, month=1, day=1,
                                hour=0, minute=0, second=0,
                                microsecond=0, tzinfo=tz)
    else:
        xdt = datetime.datetime(year=dt.year, month=12, day=31,
                                hour=23, minute=59, second=59,
                                microsecond=999999, tzinfo=tz)
    ns = (abs(dt - xdt) // datetime.timedelta(microseconds=1)) * 1000
    str_ns = str(ns)
    str_ns_stripped = str_ns if str_ns == "0" else str_ns.rstrip("0")
    lite_ns = int(str_ns_stripped)
    trailing_zeros = len(str_ns) - len(str_ns_stripped)
    tz_minutes = None
    if dt.tzinfo:
        tz_minutes = dt.utcoffset() // datetime.timedelta(minutes=1)
    return years, lite_ns, trailing_zeros, tz_minutes


def construct_datetime(year_delta, ns_delta, trailing_zeros, tz_minutes=None):
    if tz_minutes is None:
        tz = None
    else:
        tz = datetime.timezone(datetime.timedelta(minutes=tz_minutes))
    year = const.COVID_YEAR + year_delta
    ns_str = str(ns_delta) + ("0" * trailing_zeros)
    if len(ns_str) <= 3:
        ms = 0
    else:
        ms = int(ns_str[:-3])
    if year_delta >= 0:
        dt = datetime.datetime(year=year, month=1, day=1, tzinfo=tz)
    else:
        dt = datetime.datetime(year=year, month=12, day=31, hour=23,
                               minute=59, second=59, microsecond=999999,
                               tzinfo=tz)
        ms = -ms
    return dt + datetime.timedelta(microseconds=ms)


def deconstruct_date(d):
    years = d.year - const.COVID_YEAR
    if years >= 0:
        xd = datetime.date(year=d.year, month=1, day=1)
    else:
        xd = datetime.date(year=d.year, month=12, day=31)
    days = (abs(d - xd) // datetime.timedelta(days=1))
    return years, days


def construct_date(year_delta, days_delta):
    year = const.COVID_YEAR + year_delta
    if year_delta >= 0:
        dt = datetime.date(year=year, month=1, day=1)
    else:
        dt = datetime.date(year=year, month=12, day=31)
        days_delta = -days_delta
    return dt + datetime.timedelta(days=days_delta)


def deconstruct_time(t):
    val = datetime.datetime(year=const.COVID_YEAR, month=1, day=1,
                            hour=t.hour, minute=t.minute,
                            second=t.second, microsecond=t.microsecond,
                            tzinfo=t.tzinfo)
    _, ns_delta, trailing_zeros, tz = deconstruct_datetime(val)
    return ns_delta, trailing_zeros, tz


def construct_time(ns_delta, trailing_zeros, tz_minutes=None):
    dt = construct_datetime(0, ns_delta, trailing_zeros, tz_minutes=tz_minutes)
    return datetime.time(hour=dt.hour, minute=dt.minute, second=dt.second,
                         microsecond=dt.microsecond, tzinfo=dt.tzinfo)


def count_indents(line, strict=True):
    i = 0
    for char in line:
        if char == " ":
            i += 1
            continue
        break
    x = i / const.INDENT_WIDTH
    y = int(x)
    if strict and x > y:
        raise errors.IndentError
    return y


def dedent(line, indents=1):
    x = indents * const.INDENT_WIDTH
    y = 0
    for i, char in enumerate(line):
        if char == " ":
            y += 1
        else:
            break
    if y == 0:
        return line
    return line[x:] if x<= y else line[y:]


def decode_unicode(text):
    """Take in input some string that might have Unicode escape sequences.
    Output the same string with unicode escape sequences converted into
    the actual characters that they represent"""
    return text.encode("latin-1", "backslashreplace").decode("unicode-escape")


def encode_unicode(text, codec="latin-1"):  # TODO: why it isn't used ?
    """Convert a string into another where non-latin characters are
    replaced with Unicode escape sequences"""
    return text.encode(codec, "backslashreplace").decode(codec)


def get_int_base(val):
    if not isinstance(val, str):
        return 10
    parts = split_int(val)
    if parts.prefix == "0x":
        return 16
    elif parts.prefix == "0o":
        return 8
    elif parts.prefix == "0b":
        return 2
    return 10


def count_leading_zeros(val):
    parts = split_int(val)
    return len(parts.leading_zeros)


def left_pad_int(val, width):
    parts = split_int(val)
    width = 0 if width < len(parts.val) else width
    nz = width - len(parts.val) if width else 0
    leading_zeros = nz*"0"
    return parts.sign + parts.prefix + leading_zeros + parts.val


def add_leading_zeros(val, n):
    parts = split_int(val)
    leading_zeros = n * "0" if n else ""
    return parts.sign + parts.prefix + leading_zeros + parts.val


def split_int(val):
    val = val if isinstance(val, str) else str(val)
    val = val.replace("_", "")
    if val == "":
        raise errors.Error("Missing value")
    # process sign
    sign = ""
    if val.startswith("-") or val.startswith("+"):
        sign, val = val[0], val[1:]
    # process prefix
    prefix = val[0:2]
    if prefix in ("0x", "0o", "0b"):
        val = val[2:]
    else:
        prefix = ""
    # count leading zeros
    n = _count_leading_zeros(val)
    val = val.lstrip("0")
    val = "0" if not val else val
    leading_zeros = "0"*n
    return IntParts(sign, prefix, leading_zeros, val)


def split_float(s):
    """Parse a float number (string or decimal.Decimal or float), returns
    an instance of this namedtuple:
    FloatParts = namedtuple("FloatParts", ["sign", "left_significand", "right_significand", "exponent"])"""
    s = _prepare_float(s)
    # Split s at the exponent separator
    parts = s.split("E", 1)
    significand = parts[0]
    exponent = parts[1] if len(parts) == 2 else "0"
    exponent = exponent.lstrip("+")
    #exponent = (exponent[1:] if exponent and exponent[0] == "+"
    #            else exponent)
    # Split significand at the dot separator
    sign, left_significand, right_significand = _parse_significand(significand)
    # Return the named tuple
    return FloatParts(sign=sign, left_significand=left_significand,
                      right_significand=right_significand,
                      exponent=exponent)


def _count_leading_zeros(val):
    """
    This function accepts a string representing an integer.
    This string shouldn't contain a sign symbol,
    or any prefix such 0x, 0b, or 0o.
    If you want to count leading zeros for an integer representation
    that might contain a sign symbol, use the 'count_leading_zeros' function
    """
    if not val:
        return 0
    z = list()
    for x in val:
        if x == "0":
            z.append(x)
        elif x == "_":
            continue
        else:
            break
    n = len(z)
    if n == len(val):
        return n-1
    return n


def tidy_up_float(s, width=3):
    """Tidy up a float number (str or float or decimal.Decimal)
    Example: 3.141234 -> 3.141_234
    """
    s = s if isinstance(s, str) else str(s)
    # make sure that s isn't "nan", "inf", "-inf", "infinity", ...
    for char in s:
        if char not in "-01234.56789+eE":
            return s
    info = split_float(s)
    tidy_left_significand = tidy_up_int(info.left_significand)
    tidy_right_significand = ""
    if info.right_significand:
        cache = _tidy_up_right_significand(info.right_significand, width)
        tidy_right_significand = "." + cache
    tidy_exponent = "E" + info.exponent if int(info.exponent) != 0 else ""
    result = info.sign + tidy_left_significand + tidy_right_significand + tidy_exponent
    return result


def _tidy_up_right_significand(s, width):
    if not s:
        return ""
    cache = list()
    group = list()
    i = 0
    for char in s:
        if i == width:
            cache.append("".join(group))
            group = list()
            i = 0
        group.append(char)
        i += 1
    if group:
        cache.append("".join(group))
    return "_".join(cache)


def tidy_up_int(val, width=3):
    """Tidy up some int number (str or int)
    Example: 300141234 -> 300_141_234
    0xFFFFFFFF -> 0xFFFF_FFFF
    Returns a string
    """
    parts = split_int(val)
    val = parts.leading_zeros + parts.val.upper()
    tidy = list()
    group = list()
    i = 0
    for char in reversed(val):
        if i == width:
            tidy.insert(0, "".join(group))
            group = list()
            i = 0
        group.insert(0, char)
        i += 1
    if group:
        tidy.insert(0, "".join(group))
    return parts.sign + parts.prefix + "_".join(tidy)


def is_whole_number(number):
    return number % 1 == 0


def _prepare_float(s):
    if not isinstance(s, str):
        s = str(s)
    # Remove underscores, whitespaces, upper e
    cache = list()
    for c in s:
        if c in ("_", " "):
            continue
        elif c == "e":
            c = "E"
        cache.append(c)
    return "".join(cache)


def _parse_significand(s):
    if not s:
        raise errors.Error("Can't parse an empty significand")
    parts = s.split(".", 1)
    left_significand = parts[0]
    sign = left_significand[0]
    if sign in ("+", "-"):
        left_significand = left_significand[1:]
    else:
        sign = ""
    right_significand = parts[1] if len(parts) == 2 else "0"
    return sign, left_significand, right_significand



import unittest
import datetime
from paradict import errors
from paradict import serializer
from paradict.serializer.packer import Packer, pack_int
from paradict import misc, box, tags
from paradict.tags.misc import ALPHABET


class TestEmptyData(unittest.TestCase):

    def test(self):
        data = dict()
        r = pack_data(data)
        expected = misc.forge_bin(tags.DICT_EMPTY)
        self.assertEqual(expected, r)


class TestDict(unittest.TestCase):
    """
    Dicts containing complex data are tested in
    paradict.tests.test_unpacker
    """
    def test_empty_dict(self):
        data = {0: dict()}
        r = pack_data(data)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.DICT_EMPTY,
                                  tags.END)
        self.assertEqual(expected, r)


class TestList(unittest.TestCase):
    """
    Lists containing complex data are tested in
    paradict.tests.test_unpacker
    """
    def test_empty_list(self):
        data = {0: list()}
        r = pack_data(data)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.LIST_EMPTY,
                                  tags.END)
        self.assertEqual(expected, r)


class TestSet(unittest.TestCase):
    """
    Sets containing complex data are tested in
    paradict.tests.test_unpacker
    """
    def test_empty_set(self):
        data = {0: set()}
        r = pack_data(data)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.SET_EMPTY,
                                  tags.END)
        self.assertEqual(expected, r)


class TestObj(unittest.TestCase):
    """
    Objs containing complex data are tested in
    paradict.tests.test_unpacker
    """
    def test_empty_obj(self):
        data = {0: box.Obj()}
        r = pack_data(data)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.OBJ_EMPTY,
                                  tags.END)
        self.assertEqual(expected, r)


class TestGrid(unittest.TestCase):

    def test_empty_grid(self):
        data = {0: box.Grid()}
        r = pack_data(data)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.GRID_EMPTY,
                                  tags.END)
        self.assertEqual(expected, r)

    def test_with_integer(self):
        x = [(0, 1, 2),
             (3, 4, 5),
             (6, 7, -8)]
        data = {0: box.Grid(x)}
        r = pack_data(data)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.GRID,
                                  tags.CONST_0,
                                  tags.CONST_1,
                                  tags.CONST_2,
                                  tags.GRID_DIV,
                                  tags.CONST_3,
                                  tags.CONST_4,
                                  tags.CONST_5,
                                  tags.CONST_6,
                                  tags.CONST_7,
                                  tags.NINT_8, 8,
                                  tags.END,
                                  tags.END)
        self.assertEqual(expected, r)

    def test_with_float(self):
        x = [(0.0, 1.1, -2.2)]
        data = {0: box.Grid(x)}
        r = pack_data(data)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.GRID,
                                  tags.FLOAT_1,
                                  tags.CONST_0,
                                  tags.FLOAT_2,
                                  tags.CONST_1,
                                  tags.CONST_1,
                                  tags.FLOAT_2,
                                  tags.NINT_8, 2,
                                  tags.CONST_2,
                                  tags.END,
                                  tags.END)
        self.assertEqual(expected, r)

    def test_with_complex(self):
        x = [(1+2j, 3+4.5j)]
        data = {0: box.Grid(x)}
        r = pack_data(data)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.GRID,
                                  tags.COMPLEX,
                                  tags.CONST_1,
                                  tags.CONST_2,
                                  tags.COMPLEX,
                                  tags.CONST_3,
                                  tags.FLOAT_2,
                                  tags.CONST_4,
                                  tags.CONST_5,
                                  tags.END,
                                  tags.END)
        self.assertEqual(expected, r)


class TestDatetime(unittest.TestCase):

    def test_naive_datetime_lower_endpoint(self):
        year, month = 1, 1
        day, hour, minute = 1, 0, 0
        second, microsecond = 0, 0
        val = datetime.datetime(year=year, month=month, day=day,
                                hour=hour, minute=minute,
                                second=second, microsecond=microsecond)
        data = {0: val}
        r = pack_data(data)
        years, ns, trailing_zeros, _ = misc.deconstruct_datetime(val)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.DATETIME,
                                  pack_int(years),
                                  pack_int(ns),
                                  pack_int(trailing_zeros),
                                  tags.END)
        self.assertEqual(expected, r)

    def test_naive_datetime_upper_endpoint(self):
        year, month = 9999, 12
        day, hour, minute = 31, 23, 59
        second, microsecond = 59, 999999
        val = datetime.datetime(year=year, month=month, day=day,
                                hour=hour, minute=minute,
                                second=second, microsecond=microsecond)
        data = {0: val}
        r = pack_data(data)
        years, ns, trailing_zeros, _ = misc.deconstruct_datetime(val)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.DATETIME,
                                  pack_int(years),
                                  pack_int(ns),
                                  pack_int(trailing_zeros),
                                  tags.END)
        self.assertEqual(expected, r)

    def test_aware_datetime_lower_endpoint(self):
        year, month = 1, 1
        day, hour, minute = 1, 0, 0
        second, microsecond = 0, 0
        tz = datetime.timezone(datetime.timedelta(hours=0, minutes=0))
        val = datetime.datetime(year=year, month=month, day=day,
                                hour=hour, minute=minute,
                                second=second, microsecond=microsecond,
                                tzinfo=tz)
        data = {0: val}
        r = pack_data(data)
        years, ns, trailing_zeros, _ = misc.deconstruct_datetime(val)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.DATETIME_EXT,
                                  pack_int(years),
                                  pack_int(ns),
                                  pack_int(trailing_zeros),
                                  tags.CONST_0,
                                  tags.END)
        self.assertEqual(expected, r)

    def test_aware_datetime_upper_endpoint(self):
        year, month = 9999, 12
        day, hour, minute = 31, 23, 59
        second, microsecond = 59, 999999
        tz = datetime.timezone(datetime.timedelta(hours=23, minutes=59))
        val = datetime.datetime(year=year, month=month, day=day,
                                hour=hour, minute=minute,
                                second=second, microsecond=microsecond,
                                tzinfo=tz)
        data = {0: val}
        r = pack_data(data)
        years, ns, trailing_zeros, tz_minutes = misc.deconstruct_datetime(val)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.DATETIME_EXT,
                                  pack_int(years),
                                  pack_int(ns),
                                  pack_int(trailing_zeros),
                                  pack_int(tz_minutes),
                                  tags.END)
        self.assertEqual(expected, r)

    def test_aware_datetime_with_negative_negative_utc_timezone(self):
        year, month = 9999, 12
        day, hour, minute = 31, 23, 59
        second, microsecond = 59, 999999
        tz = datetime.timezone(datetime.timedelta(hours=-1, minutes=-30))
        val = datetime.datetime(year=year, month=month, day=day,
                                hour=hour, minute=minute,
                                second=second, microsecond=microsecond,
                                tzinfo=tz)
        data = {0: val}
        r = pack_data(data)
        years, ns, trailing_zeros, tz_minutes = misc.deconstruct_datetime(val)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.DATETIME_EXT,
                                  pack_int(years),
                                  pack_int(ns),
                                  pack_int(trailing_zeros),
                                  pack_int(tz_minutes),
                                  tags.END)
        self.assertEqual(expected, r)


class TestDate(unittest.TestCase):

    def test_date_lower_endpoint(self):
        year, month, day = 1, 1, 1
        val = datetime.date(year=year, month=month, day=day)
        data = {0: val}
        r = pack_data(data)
        years, days = misc.deconstruct_date(val)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.DATE,
                                  pack_int(years),
                                  pack_int(days),
                                  tags.END)
        self.assertEqual(expected, r)

    def test_date_upper_endpoint(self):
        year, month, day = 9999, 12, 31
        val = datetime.date(year=year, month=month, day=day)
        data = {0: val}
        r = pack_data(data)
        years, days = misc.deconstruct_date(val)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.DATE,
                                  pack_int(years),
                                  pack_int(days),
                                  tags.END)
        self.assertEqual(expected, r)


class TestTime(unittest.TestCase):

    def test_naive_time_lower_endpoint(self):
        hour, minute = 0, 0
        second, microsecond = 0, 0
        val = datetime.time(hour=hour, minute=minute,
                            second=second, microsecond=microsecond)
        data = {0: val}
        r = pack_data(data)
        ns, trailing_zeros, _ = misc.deconstruct_time(val)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.TIME,
                                  pack_int(ns),
                                  pack_int(trailing_zeros),
                                  tags.END)
        self.assertEqual(expected, r)

    def test_naive_time_upper_endpoint(self):
        hour, minute = 23, 59
        second, microsecond = 59, 999999
        val = datetime.time(hour=hour, minute=minute,
                            second=second, microsecond=microsecond)
        data = {0: val}
        r = pack_data(data)
        ns, trailing_zeros, _ = misc.deconstruct_time(val)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.TIME,
                                  pack_int(ns),
                                  pack_int(trailing_zeros),
                                  tags.END)
        self.assertEqual(expected, r)

    def test_aware_time_lower_endpoint(self):
        hour, minute = 0, 0
        second, microsecond = 0, 0
        tz = datetime.timezone(datetime.timedelta(hours=0, minutes=0))
        val = datetime.time(hour=hour, minute=minute,
                            second=second, microsecond=microsecond,
                            tzinfo=tz)
        data = {0: val}
        r = pack_data(data)
        ns, trailing_zeros, _ = misc.deconstruct_time(val)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.TIME_EXT,
                                  pack_int(ns),
                                  pack_int(trailing_zeros),
                                  tags.CONST_0,
                                  tags.END)
        self.assertEqual(expected, r)

    def test_aware_time_upper_endpoint(self):
        hour, minute = 23, 59
        second, microsecond = 59, 999999
        tz = datetime.timezone(datetime.timedelta(hours=0, minutes=0))
        val = datetime.time(hour=hour, minute=minute,
                            second=second, microsecond=microsecond,
                            tzinfo=tz)
        data = {0: val}
        r = pack_data(data)
        ns, trailing_zeros, tz_minutes = misc.deconstruct_time(val)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.TIME_EXT,
                                  pack_int(ns),
                                  pack_int(trailing_zeros),
                                  pack_int(tz_minutes),
                                  tags.END)
        self.assertEqual(expected, r)

    def test_aware_datetime_with_negative_negative_utc_timezone(self):
        hour, minute = 23, 59
        second, microsecond = 59, 999999
        tz = datetime.timezone(datetime.timedelta(hours=-1, minutes=-30))
        val = datetime.time(hour=hour, minute=minute,
                            second=second, microsecond=microsecond,
                            tzinfo=tz)
        data = {0: val}
        r = pack_data(data)
        ns, trailing_zeros, tz_minutes = misc.deconstruct_time(val)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.TIME_EXT,
                                  pack_int(ns),
                                  pack_int(trailing_zeros),
                                  pack_int(tz_minutes),
                                  tags.END)
        self.assertEqual(expected, r)


class TestNull(unittest.TestCase):

    def test(self):
        data = {0: None}
        r = pack_data(data)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.NULL,
                                  tags.END)
        self.assertEqual(expected, r)


class TestBool(unittest.TestCase):

    def test_true(self):
        data = {0: True}
        r = pack_data(data)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.BOOL_TRUE,
                                  tags.END)
        self.assertEqual(expected, r)

    def test_false(self):
        data = {0: False}
        r = pack_data(data)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.BOOL_FALSE,
                                  tags.END)
        self.assertEqual(expected, r)


class TestConstChar(unittest.TestCase):

    def test_lower_letters(self):
        letters = ["a", "b", "c", "d", "e", "f", "g", "h", "i", "j",
                   "k", "l", "m", "n", "o", "p", "q", "r", "s", "t",
                   "u", "v", "w", "x", "y", "z"]
        for letter in letters:
            with self.subTest("Letter: {}".format(letter)):
                data = {letter: letter}
                packer = Packer(data)
                r = pack_data(data)
                const_tag = ALPHABET.get(letter)
                expected = misc.forge_bin(tags.DICT,
                                          const_tag, const_tag,
                                          tags.END)
                self.assertEqual(expected, r)

    def test_upper_letters(self):
        letters = ["A", "B", "C", "D", "E", "F", "G", "H", "I", "J",
                   "K", "L", "M", "N", "O", "P", "Q", "R", "S", "T",
                   "U", "V", "W", "X", "Y", "Z"]
        for letter in letters:
            with self.subTest("Letter: {}".format(letter)):
                data = {letter: letter}
                packer = Packer(data)
                r = pack_data(data)
                const_tag = ALPHABET.get(letter)
                expected = misc.forge_bin(tags.DICT,
                                          const_tag, const_tag,
                                          tags.END)
                self.assertEqual(expected, r)


class TestString(unittest.TestCase):

    def test_empty_str(self):
        data = {0: ""}
        r = pack_data(data)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.STR_EMPTY,
                                  tags.END)
        self.assertEqual(expected, r)

    def test_str_8(self):
        x = ";"
        bin_x = x.encode("utf-8")
        data = {0: x}
        r = pack_data(data)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.STR_8, bin_x,
                                  tags.END)
        self.assertEqual(expected, r)

    def test_str_16(self):
        x = ";"*2
        bin_x = x.encode("utf-8")
        data = {0: x}
        r = pack_data(data)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.STR_16, bin_x,
                                  tags.END)
        self.assertEqual(expected, r)

    def test_str_24(self):
        x = ";"*3
        bin_x = x.encode("utf-8")
        data = {0: x}
        r = pack_data(data)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.STR_24, bin_x,
                                  tags.END)
        self.assertEqual(expected, r)

    def test_str_32(self):
        x = ";"*4
        bin_x = x.encode("utf-8")
        data = {0: x}
        r = pack_data(data)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.STR_32, bin_x,
                                  tags.END)
        self.assertEqual(expected, r)

    def test_str_40(self):
        x = ";"*5
        bin_x = x.encode("utf-8")
        data = {0: x}
        r = pack_data(data)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.STR_40, bin_x,
                                  tags.END)
        self.assertEqual(expected, r)

    def test_str_48(self):
        x = ";"*6
        bin_x = x.encode("utf-8")
        data = {0: x}
        r = pack_data(data)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.STR_48, bin_x,
                                  tags.END)
        self.assertEqual(expected, r)

    def test_str_56(self):
        x = ";"*7
        bin_x = x.encode("utf-8")
        data = {0: x}
        r = pack_data(data)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.STR_56, bin_x,
                                  tags.END)
        self.assertEqual(expected, r)

    def test_str_64(self):
        x = ";"*8
        bin_x = x.encode("utf-8")
        data = {0: x}
        r = pack_data(data)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.STR_64, bin_x,
                                  tags.END)
        self.assertEqual(expected, r)

    def test_str_72(self):
        x = ";"*9
        bin_x = x.encode("utf-8")
        data = {0: x}
        r = pack_data(data)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.STR_72, bin_x,
                                  tags.END)
        self.assertEqual(expected, r)

    def test_str_80(self):
        x = ";"*10
        bin_x = x.encode("utf-8")
        data = {0: x}
        r = pack_data(data)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.STR_80, bin_x,
                                  tags.END)
        self.assertEqual(expected, r)

    def test_str_88(self):
        x = ";"*11
        bin_x = x.encode("utf-8")
        data = {0: x}
        r = pack_data(data)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.STR_88, bin_x,
                                  tags.END)
        self.assertEqual(expected, r)

    def test_str_96(self):
        x = ";"*12
        bin_x = x.encode("utf-8")
        data = {0: x}
        r = pack_data(data)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.STR_96, bin_x,
                                  tags.END)
        self.assertEqual(expected, r)

    def test_str_104(self):
        x = ";"*13
        bin_x = x.encode("utf-8")
        data = {0: x}
        r = pack_data(data)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.STR_104, bin_x,
                                  tags.END)
        self.assertEqual(expected, r)

    def test_str_112(self):
        x = ";"*14
        bin_x = x.encode("utf-8")
        data = {0: x}
        r = pack_data(data)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.STR_112, bin_x,
                                  tags.END)
        self.assertEqual(expected, r)

    def test_str_120(self):
        x = ";"*15
        bin_x = x.encode("utf-8")
        data = {0: x}
        r = pack_data(data)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.STR_120, bin_x,
                                  tags.END)
        self.assertEqual(expected, r)

    def test_str_128(self):
        x = ";"*16
        bin_x = x.encode("utf-8")
        data = {0: x}
        r = pack_data(data)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.STR_128, bin_x,
                                  tags.END)
        self.assertEqual(expected, r)

    def test_str_136(self):
        x = ";"*17
        bin_x = x.encode("utf-8")
        data = {0: x}
        r = pack_data(data)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.STR_136, bin_x,
                                  tags.END)
        self.assertEqual(expected, r)

    def test_str_144(self):
        x = ";"*18
        bin_x = x.encode("utf-8")
        data = {0: x}
        r = pack_data(data)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.STR_144, bin_x,
                                  tags.END)
        self.assertEqual(expected, r)

    def test_str_152(self):
        x = ";"*19
        bin_x = x.encode("utf-8")
        data = {0: x}
        r = pack_data(data)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.STR_152, bin_x,
                                  tags.END)
        self.assertEqual(expected, r)

    def test_str_160(self):
        x = ";"*20
        bin_x = x.encode("utf-8")
        data = {0: x}
        r = pack_data(data)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.STR_160, bin_x,
                                  tags.END)
        self.assertEqual(expected, r)

    def test_str_168(self):
        x = ";"*21
        bin_x = x.encode("utf-8")
        data = {0: x}
        r = pack_data(data)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.STR_168, bin_x,
                                  tags.END)
        self.assertEqual(expected, r)

    def test_str_176(self):
        x = ";"*22
        bin_x = x.encode("utf-8")
        data = {0: x}
        r = pack_data(data)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.STR_176, bin_x,
                                  tags.END)
        self.assertEqual(expected, r)

    def test_str_184(self):
        x = ";"*23
        bin_x = x.encode("utf-8")
        data = {0: x}
        r = pack_data(data)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.STR_184, bin_x,
                                  tags.END)
        self.assertEqual(expected, r)

    def test_str_192(self):
        x = ";"*24
        bin_x = x.encode("utf-8")
        data = {0: x}
        r = pack_data(data)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.STR_192, bin_x,
                                  tags.END)
        self.assertEqual(expected, r)

    def test_str_200(self):
        x = ";"*25
        bin_x = x.encode("utf-8")
        data = {0: x}
        r = pack_data(data)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.STR_200, bin_x,
                                  tags.END)
        self.assertEqual(expected, r)

    def test_str_208(self):
        x = ";"*26
        bin_x = x.encode("utf-8")
        data = {0: x}
        r = pack_data(data)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.STR_208, bin_x,
                                  tags.END)
        self.assertEqual(expected, r)

    def test_str_216(self):
        x = ";"*27
        bin_x = x.encode("utf-8")
        data = {0: x}
        r = pack_data(data)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.STR_216, bin_x,
                                  tags.END)
        self.assertEqual(expected, r)

    def test_str_224(self):
        x = ";"*28
        bin_x = x.encode("utf-8")
        data = {0: x}
        r = pack_data(data)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.STR_224, bin_x,
                                  tags.END)
        self.assertEqual(expected, r)

    def test_str_232(self):
        x = ";"*29
        bin_x = x.encode("utf-8")
        data = {0: x}
        r = pack_data(data)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.STR_232, bin_x,
                                  tags.END)
        self.assertEqual(expected, r)

    def test_str_240(self):
        x = ";"*30
        bin_x = x.encode("utf-8")
        data = {0: x}
        r = pack_data(data)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.STR_240, bin_x,
                                  tags.END)
        self.assertEqual(expected, r)

    def test_str_248(self):
        x = ";"*31
        bin_x = x.encode("utf-8")
        data = {0: x}
        r = pack_data(data)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.STR_248, bin_x,
                                  tags.END)
        self.assertEqual(expected, r)

    def test_str_256(self):
        x = ";"*32
        bin_x = x.encode("utf-8")
        data = {0: x}
        r = pack_data(data)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.STR_256, bin_x,
                                  tags.END)
        self.assertEqual(expected, r)

    def test_short_str_lower_endpoint(self):
        x = ";"*33
        bin_x = x.encode("utf-8")
        data = {0: x}
        r = pack_data(data)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.STR_SHORT, len(bin_x) - 1, bin_x,
                                  tags.END)
        self.assertEqual(expected, r)

    def test_short_str_upper_endpoint(self):
        x = ";"*(2**8)
        bin_x = x.encode("utf-8")
        data = {0: x}
        r = pack_data(data)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.STR_SHORT, len(bin_x) - 1, bin_x,
                                  tags.END)
        self.assertEqual(expected, r)

    def test_medium_str_lower_endpoint(self):
        x = ";" * (2**8 + 1)
        bin_x = x.encode("utf-8")
        data = {0: x}
        r = pack_data(data)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.STR_MEDIUM, len(bin_x) - 1, bin_x,
                                  tags.END)
        self.assertEqual(expected, r)

    def test_medium_str_upper_endpoint(self):
        x = ";" * (2**16)
        bin_x = x.encode("utf-8")
        data = {0: x}
        r = pack_data(data)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.STR_MEDIUM, len(bin_x) - 1, bin_x,
                                  tags.END)
        self.assertEqual(expected, r)

    def test_long_str_lower_endpoint(self):
        x = ";" * (2**16 + 1)
        bin_x = x.encode("utf-8")
        data = {0: x}
        r = pack_data(data)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.STR_LONG, len(bin_x) - 1, bin_x,
                                  tags.END)
        self.assertEqual(expected, r)

    def test_long_str_upper(self):
        x = ";" * (2**24)
        bin_x = x.encode("utf-8")
        data = {0: x}
        r = pack_data(data)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.STR_LONG, len(bin_x) - 1, bin_x,
                                  tags.END)
        self.assertEqual(expected, r)

    def test_big_str(self):
        # toooooo big to test ;)
        self.assertTrue(True)

    def test_heavy_str(self):
        # toooooo heavy to test ;)
        self.assertTrue(True)


class TestComment(unittest.TestCase):

    def test_empty_comment(self):
        data = {box.CommentID(): box.Comment()}
        r = pack_data(data, skip_comments=False)
        expected = misc.forge_bin(tags.DICT,
                                  tags.COMMENT,
                                  tags.NULL,
                                  tags.COMMENT,
                                  tags.STR_EMPTY,
                                  tags.END)
        self.assertEqual(expected, r)

    def test_skipped_comment(self):
        data = {box.CommentID(): box.Comment()}
        r = pack_data(data, skip_comments=True)
        expected = misc.forge_bin(tags.DICT,
                                  tags.END)
        self.assertEqual(expected, r)

    def test_one_char_comment(self):
        x = box.Comment("#")
        bin_x = x.encode("utf-8")
        data = {box.CommentID(): x}
        r = pack_data(data, skip_comments=False)
        expected = misc.forge_bin(tags.DICT,
                                  tags.COMMENT,
                                  tags.NULL,
                                  tags.COMMENT,
                                  tags.STR_8, bin_x,
                                  tags.END)
        self.assertEqual(expected, r)

    def test_two_char_comment(self):
        x = box.Comment("##")
        bin_x = x.encode("utf-8")
        data = {box.CommentID(): x}
        r = pack_data(data, skip_comments=False)
        expected = misc.forge_bin(tags.DICT,
                                  tags.COMMENT,
                                  tags.NULL,
                                  tags.COMMENT,
                                  tags.STR_16, bin_x,
                                  tags.END)
        self.assertEqual(expected, r)

    def test_short_comment(self):
        x = box.Comment("#"*(2**8))
        bin_x = x.encode("utf-8")
        data = {box.CommentID(): x}
        r = pack_data(data, skip_comments=False)
        expected = misc.forge_bin(tags.DICT,
                                  tags.COMMENT,
                                  tags.NULL,
                                  tags.COMMENT,
                                  tags.STR_SHORT, len(bin_x) - 1, bin_x,
                                  tags.END)
        self.assertEqual(expected, r)

    def test_long_comment(self):
        x = box.Comment("#" * (2**24))
        bin_x = x.encode("utf-8")
        data = {box.CommentID(): x}
        r = pack_data(data, skip_comments=False)
        expected = misc.forge_bin(tags.DICT,
                                  tags.COMMENT,
                                  tags.NULL,
                                  tags.COMMENT,
                                  tags.STR_LONG, len(bin_x) - 1, bin_x,
                                  tags.END)
        self.assertEqual(expected, r)


class TestBin(unittest.TestCase):

    def test_empty_bin(self):
        data = {0: b''}
        r = pack_data(data)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.BIN_EMPTY,
                                  tags.END)
        self.assertEqual(expected, r)

    def test_short_bin_lower_endpoint(self):
        x = b'\x00'*1
        data = {0: x}
        r = pack_data(data)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.BIN_SHORT, len(x) - 1, x,
                                  tags.END)
        self.assertEqual(expected, r)

    def test_short_bin_upper_endpoint(self):
        x = b'\x00'*(2**8)
        data = {0: x}
        r = pack_data(data)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.BIN_SHORT, len(x) - 1, x,
                                  tags.END)
        self.assertEqual(expected, r)

    def test_medium_bin_lower_endpoint(self):
        x = b'\x00' * (2**8 + 1)
        data = {0: x}
        r = pack_data(data)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.BIN_MEDIUM, len(x) - 1, x,
                                  tags.END)
        self.assertEqual(expected, r)

    def test_medium_bin_upper_endpoint(self):
        x = b'\x00' * (2**16)
        data = {0: x}
        r = pack_data(data)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.BIN_MEDIUM, len(x) - 1, x,
                                  tags.END)
        self.assertEqual(expected, r)

    def test_long_bin_lower_endpoint(self):
        x = b'\x00' * (2**16 + 1)
        data = {0: x}
        r = pack_data(data)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.BIN_LONG, len(x) - 1, x,
                                  tags.END)
        self.assertEqual(expected, r)

    def test_long_bin_upper(self):
        x = b'\x00' * (2**24)
        data = {0: x}
        r = pack_data(data)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.BIN_LONG, len(x) - 1, x,
                                  tags.END)
        self.assertEqual(expected, r)

    def test_big_bin(self):
        # toooooo big to test ;)
        self.assertTrue(True)

    def test_heavy_bin(self):
        # toooooo heavy to test ;)
        self.assertTrue(True)


class TestConstInt(unittest.TestCase):

    def test_const_number_0(self):
        data = {0: 0}
        r = pack_data(data)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.CONST_0,
                                  tags.END)
        self.assertEqual(expected, r)

    def test_const_number_1(self):
        data = {1: 1}
        r = pack_data(data)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_1,
                                  tags.CONST_1,
                                  tags.END)
        self.assertEqual(expected, r)

    def test_const_number_99(self):
        data = {99: 99}
        r = pack_data(data)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_99,
                                  tags.CONST_99,
                                  tags.END)
        self.assertEqual(expected, r)


class TestPositiveInt(unittest.TestCase):

    def test_pint8_lower(self):
        x = 100
        data = {0: x}
        r = pack_data(data)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.PINT_8, x,
                                  tags.END)
        self.assertEqual(expected, r)

    def test_pint8_upper(self):
        x = 2**8 - 1
        data = {0: x}
        r = pack_data(data)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.PINT_8, x,
                                  tags.END)
        self.assertEqual(expected, r)

    def test_pint16_lower(self):
        x = 2**8
        data = {0: x}
        r = pack_data(data)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.PINT_16, x,
                                  tags.END)
        self.assertEqual(expected, r)

    def test_pint16_upper(self):
        x = 2**16 - 1
        data = {0: x}
        r = pack_data(data)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.PINT_16, x,
                                  tags.END)
        self.assertEqual(expected, r)

    def test_pint24_lower(self):
        x = 2**16
        data = {0: x}
        r = pack_data(data)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.PINT_24, x,
                                  tags.END)
        self.assertEqual(expected, r)

    def test_pint24_upper(self):
        x = 2**24 - 1
        data = {0: x}
        r = pack_data(data)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.PINT_24, x,
                                  tags.END)
        self.assertEqual(expected, r)

    def test_pint32_lower(self):
        x = 2**24
        data = {0: x}
        r = pack_data(data)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.PINT_32, x,
                                  tags.END)
        self.assertEqual(expected, r)

    def test_pint32_upper(self):
        x = 2**32 - 1
        data = {0: x}
        r = pack_data(data)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.PINT_32, x,
                                  tags.END)
        self.assertEqual(expected, r)

    def test_pint40_lower(self):
        x = 2**32
        data = {0: x}
        r = pack_data(data)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.PINT_40, x,
                                  tags.END)
        self.assertEqual(expected, r)

    def test_pint40_upper(self):
        x = 2**40 - 1
        data = {0: x}
        r = pack_data(data)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.PINT_40, x,
                                  tags.END)
        self.assertEqual(expected, r)

    def test_pint48_lower(self):
        x = 2**40
        data = {0: x}
        r = pack_data(data)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.PINT_48, x,
                                  tags.END)
        self.assertEqual(expected, r)

    def test_pint48_upper(self):
        x = 2**48 - 1
        data = {0: x}
        r = pack_data(data)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.PINT_48, x,
                                  tags.END)
        self.assertEqual(expected, r)

    def test_pint56_lower(self):
        x = 2**48
        data = {0: x}
        r = pack_data(data)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.PINT_56, x,
                                  tags.END)
        self.assertEqual(expected, r)

    def test_pint56_upper(self):
        x = 2**56 - 1
        data = {0: x}
        r = pack_data(data)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.PINT_56, x,
                                  tags.END)
        self.assertEqual(expected, r)

    def test_pint64_lower(self):
        x = 2**56
        data = {0: x}
        r = pack_data(data)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.PINT_64, x,
                                  tags.END)
        self.assertEqual(expected, r)

    def test_pint64_upper(self):
        x = 2**64 - 1
        data = {0: x}
        r = pack_data(data)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.PINT_64, x,
                                  tags.END)
        self.assertEqual(expected, r)

    def test_pint_big_lower(self):
        x = 2**64
        data = {0: x}
        size = misc.calc_uint_bytes(x)
        r = pack_data(data)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.PINT_BIG, size - 1, x,
                                  tags.END)
        self.assertEqual(expected, r)

    def test_pint_big_upper(self):
        x = 2**2048 - 1
        data = {0: x}
        size = misc.calc_uint_bytes(x)  # 256 bytes !
        r = pack_data(data)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.PINT_BIG, size - 1, x,
                                  tags.END)
        self.assertEqual(expected, r)

    def test_pint_heavy_lower(self):
        x = 2**2048
        data = {0: x}
        size = misc.calc_uint_bytes(x)
        r = pack_data(data)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.PINT_HEAVY, size - 1, x,
                                  tags.END)
        self.assertEqual(expected, r)

    def test_pint_heavy_upper(self):
        x = 2**524288 - 1
        data = {0: x}
        size = misc.calc_uint_bytes(x)
        r = pack_data(data)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.PINT_HEAVY, size - 1, x,
                                  tags.END)
        self.assertEqual(expected, r)

    def test_too_large_number(self):
        x = 2**524289
        data = {0: x}
        packer = Packer(data)
        with self.assertRaises(errors.Error):
            pack_data(packer)


class TestNegativeInt(unittest.TestCase):

    def test_nint8_lower(self):
        x = 100
        data = {0: -x}
        r = pack_data(data)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.NINT_8, x,
                                  tags.END)
        self.assertEqual(expected, r)

    def test_nint8_upper(self):
        x = 2**8 - 1
        data = {0:- x}
        r = pack_data(data)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.NINT_8, x,
                                  tags.END)
        self.assertEqual(expected, r)

    def test_nint16_lower(self):
        x = 2**8
        data = {0: -x}
        r = pack_data(data)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.NINT_16, x,
                                  tags.END)
        self.assertEqual(expected, r)

    def test_nint16_upper(self):
        x = 2**16 - 1
        data = {0: -x}
        r = pack_data(data)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.NINT_16, x,
                                  tags.END)
        self.assertEqual(expected, r)

    def test_nint24_lower(self):
        x = 2**16
        data = {0: -x}
        r = pack_data(data)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.NINT_24, x,
                                  tags.END)
        self.assertEqual(expected, r)

    def test_nint24_upper(self):
        x = 2**24 - 1
        data = {0: -x}
        r = pack_data(data)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.NINT_24, x,
                                  tags.END)
        self.assertEqual(expected, r)

    def test_nint32_lower(self):
        x = 2**24
        data = {0: -x}
        r = pack_data(data)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.NINT_32, x,
                                  tags.END)
        self.assertEqual(expected, r)

    def test_nint32_upper(self):
        x = 2**32 - 1
        data = {0: -x}
        r = pack_data(data)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.NINT_32, x,
                                  tags.END)
        self.assertEqual(expected, r)

    def test_nint40_lower(self):
        x = 2**32
        data = {0: -x}
        r = pack_data(data)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.NINT_40, x,
                                  tags.END)
        self.assertEqual(expected, r)

    def test_nint40_upper(self):
        x = 2**40 - 1
        data = {0: -x}
        r = pack_data(data)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.NINT_40, x,
                                  tags.END)
        self.assertEqual(expected, r)

    def test_nint48_lower(self):
        x = 2**40
        data = {0: -x}
        r = pack_data(data)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.NINT_48, x,
                                  tags.END)
        self.assertEqual(expected, r)

    def test_nint48_upper(self):
        x = 2**48 - 1
        data = {0: -x}
        r = pack_data(data)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.NINT_48, x,
                                  tags.END)
        self.assertEqual(expected, r)

    def test_nint56_lower(self):
        x = 2**48
        data = {0: -x}
        r = pack_data(data)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.NINT_56, x,
                                  tags.END)
        self.assertEqual(expected, r)

    def test_nint56_upper(self):
        x = 2**56 - 1
        data = {0: -x}
        r = pack_data(data)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.NINT_56, x,
                                  tags.END)
        self.assertEqual(expected, r)

    def test_nint64_lower(self):
        x = 2**56
        data = {0: -x}
        r = pack_data(data)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.NINT_64, x,
                                  tags.END)
        self.assertEqual(expected, r)

    def test_nint64_upper(self):
        x = 2**64 - 1
        data = {0: -x}
        r = pack_data(data)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.NINT_64, x,
                                  tags.END)
        self.assertEqual(expected, r)

    def test_nint_big_lower(self):
        x = 2**64
        data = {0: -x}
        size = misc.calc_uint_bytes(x)
        r = pack_data(data)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.NINT_BIG, size - 1, x,
                                  tags.END)
        self.assertEqual(expected, r)

    def test_nint_big_upper(self):
        x = 2**2048 - 1
        data = {0: -x}
        size = misc.calc_uint_bytes(x)  # 256 bytes !
        r = pack_data(data)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.NINT_BIG, size - 1, x,
                                  tags.END)
        self.assertEqual(expected, r)

    def test_nint_heavy_lower(self):
        x = 2**2048
        data = {0: -x}
        size = misc.calc_uint_bytes(x)
        r = pack_data(data)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.NINT_HEAVY, size - 1, x,
                                  tags.END)
        self.assertEqual(expected, r)

    def test_nint_heavy_upper(self):
        x = 2**524288 - 1
        data = {0: -x}
        size = misc.calc_uint_bytes(x)
        r = pack_data(data)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.NINT_HEAVY, size - 1, x,
                                  tags.END)
        self.assertEqual(expected, r)

    def test_too_large_number(self):
        x = 2**524289
        data = {0: -x}
        packer = Packer(data)
        with self.assertRaises(errors.Error):
            pack_data(packer)


class TestIntAsHexOctBin(unittest.TestCase):

    def test_zero_as_hex(self):
        x = box.HexInt(0)
        data = {0: x}
        r = pack_data(data)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.RADIX_HEX,
                                  tags.CONST_0,
                                  tags.END)
        self.assertEqual(expected, r)

    def test_zero_as_oct(self):
        x = box.OctInt(0)
        data = {0: x}
        r = pack_data(data)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.RADIX_OCT,
                                  tags.CONST_0,
                                  tags.END)
        self.assertEqual(expected, r)

    def test_zero_as_bin(self):
        x = box.BinInt(0)
        data = {0: x}
        r = pack_data(data)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.RADIX_BIN,
                                  tags.CONST_0,
                                  tags.END)
        self.assertEqual(expected, r)

    def test_pint_as_hex(self):
        val = 2**8 - 1
        x = box.HexInt(val)
        data = {0: x}
        r = pack_data(data)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.RADIX_HEX,
                                  pack_int(val),
                                  tags.END)
        self.assertEqual(expected, r)

    def test_pint_as_oct(self):
        val = 2**8 - 1
        x = box.OctInt(val)
        data = {0: x}
        r = pack_data(data)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.RADIX_OCT,
                                  pack_int(val),
                                  tags.END)
        self.assertEqual(expected, r)

    def test_pint_as_bin(self):
        val = 2**8 - 1
        x = box.BinInt(val)
        data = {0: x}
        r = pack_data(data)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.RADIX_BIN,
                                  pack_int(val),
                                  tags.END)
        self.assertEqual(expected, r)

    def test_nint_as_hex(self):
        val = 2**8 - 1
        x = box.HexInt(-val)
        data = {0: x}
        r = pack_data(data)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.RADIX_HEX,
                                  pack_int(-val),
                                  tags.END)
        self.assertEqual(expected, r)

    def test_nint_as_oct(self):
        val = 2**8 - 1
        x = box.OctInt(-val)
        data = {0: x}
        r = pack_data(data)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.RADIX_OCT,
                                  pack_int(-val),
                                  tags.END)
        self.assertEqual(expected, r)

    def test_nint_as_bin(self):
        val = 2**8 - 1
        x = box.BinInt(-val)
        data = {0: x}
        r = pack_data(data)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.RADIX_BIN,
                                  pack_int(-val),
                                  tags.END)
        self.assertEqual(expected, r)


class TestIntAsStringHexOctBin(unittest.TestCase):

    def test_zero_as_hex(self):
        x = box.HexInt("0x0000")
        data = {0: x}
        r = pack_data(data)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.RADIX_HEX_EXT,
                                  tags.CONST_3,
                                  tags.CONST_0,
                                  tags.END)
        self.assertEqual(expected, r)

    def test_zero_as_oct(self):
        x = box.OctInt("0o0000")
        data = {0: x}
        r = pack_data(data)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.RADIX_OCT_EXT,
                                  tags.CONST_3,
                                  tags.CONST_0,
                                  tags.END)
        self.assertEqual(expected, r)

    def test_zero_as_bin(self):
        x = box.BinInt("0b0000")
        data = {0: x}
        r = pack_data(data)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.RADIX_BIN_EXT,
                                  tags.CONST_3,
                                  tags.CONST_0,
                                  tags.END)
        self.assertEqual(expected, r)

    def test_pint_as_hex(self):
        val = 2**8 - 1
        x = box.HexInt("0x0000_00ff")
        data = {0: x}
        r = pack_data(data)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.RADIX_HEX_EXT,
                                  tags.CONST_6,
                                  pack_int(val),
                                  tags.END)
        self.assertEqual(expected, r)

    def test_pint_as_oct(self):
        val = 2**8 - 1
        x = box.OctInt("0o00_000_377")
        data = {0: x}
        r = pack_data(data)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.RADIX_OCT_EXT,
                                  tags.CONST_5,
                                  pack_int(val),
                                  tags.END)
        self.assertEqual(expected, r)

    def test_pint_as_bin(self):
        val = 2**8 - 1
        x = box.BinInt("0b1111_1111")
        data = {0: x}
        r = pack_data(data)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.RADIX_BIN,
                                  pack_int(val),
                                  tags.END)
        self.assertEqual(expected, r)

    def test_nint_as_hex(self):
        val = 2**8 - 1
        x = box.HexInt("-0x0000_00ff")
        data = {0: x}
        r = pack_data(data)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.RADIX_HEX_EXT,
                                  tags.CONST_6,
                                  pack_int(-val),
                                  tags.END)
        self.assertEqual(expected, r)

    def test_nint_as_oct(self):
        val = 2**8 - 1
        x = box.OctInt("-0o00_000_377")
        data = {0: x}
        r = pack_data(data)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.RADIX_OCT_EXT,
                                  tags.CONST_5,
                                  pack_int(-val),
                                  tags.END)
        self.assertEqual(expected, r)

    def test_nint_as_bin(self):
        val = 2**8 - 1
        x = box.BinInt("-0b11111111")
        data = {0: x}
        r = pack_data(data)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.RADIX_BIN,
                                  pack_int(-val),
                                  tags.END)
        self.assertEqual(expected, r)


class TestFloatNumber(unittest.TestCase):

    def test_zero(self):
        x = 0.0
        data = {0: x}
        r = pack_data(data)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.FLOAT_1,
                                  tags.CONST_0,
                                  tags.END)
        self.assertEqual(expected, r)

    def test_negative_zero(self):
        x = float("-0.0")
        data = {0: x}
        r = pack_data(data)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.FLOAT_MISC,
                                  tags.CHAR_Z,
                                  tags.END)
        self.assertEqual(expected, r)

    def test_positive_infinity(self):
        x = float("+inf")
        data = {0: x}
        r = pack_data(data)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.FLOAT_MISC,
                                  tags.CHAR_X,
                                  tags.END)
        self.assertEqual(expected, r)

    def test_negative_infinity(self):
        x = float("-inf")
        data = {0: x}
        r = pack_data(data)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.FLOAT_MISC,
                                  tags.CHAR_Y,
                                  tags.END)
        self.assertEqual(expected, r)

    def test_positive_float_1(self):
        x = float(1)
        data = {0: x}
        r = pack_data(data)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.FLOAT_1,
                                  tags.CONST_1,
                                  tags.END)
        self.assertEqual(expected, r)

    def test_positive_float_1_ext(self):
        x = float("1E-10")
        data = {0: x}
        r = pack_data(data)
        parts = misc.split_float(x)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.FLOAT_1_EXT,
                                  tags.CONST_1,
                                  tags.NINT_8, abs(int(parts.exponent)),
                                  tags.END)
        self.assertEqual(expected, r)

    def test_negative_float_1(self):
        x = float(1)
        data = {0: -x}
        r = pack_data(data)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.FLOAT_1,
                                  tags.NINT_8, 1,
                                  tags.END)
        self.assertEqual(expected, r)

    def test_negative_float_1_ext(self):
        x = float("1E-10")
        data = {0: -x}
        r = pack_data(data)
        parts = misc.split_float(x)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.FLOAT_1_EXT,
                                  tags.NINT_8, 1,
                                  tags.NINT_8, abs(int(parts.exponent)),
                                  tags.END)
        self.assertEqual(expected, r)

    def test_positive_float_2(self):
        left_significand = 1
        right_significand = 2 ** 32 - 1
        significand = "{}.{}".format(left_significand, right_significand)
        x = float(significand)
        data = {0: x}
        r = pack_data(data)
        parts = misc.split_float(x)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.FLOAT_2,
                                  tags.CONST_1,
                                  tags.PINT_32, int(parts.right_significand),
                                  tags.END)
        self.assertEqual(expected, r)

    def test_positive_float_2_ext(self):
        left_significand = 1
        right_significand = 2**32 - 1

        significand = "{}.{}".format(left_significand, right_significand)
        x = float("{}E-10".format(significand))
        data = {0: x}
        r = pack_data(data)
        parts = misc.split_float(x)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.FLOAT_2_EXT,
                                  tags.CONST_1,
                                  tags.PINT_32, int(parts.right_significand),
                                  tags.NINT_8, abs(int(parts.exponent)),
                                  tags.END)
        self.assertEqual(expected, r)

    def test_negative_float_2(self):
        left_significand = 1
        right_significand = 2 ** 32 - 1
        significand = "{}.{}".format(left_significand, right_significand)
        x = float(significand)
        data = {0: -x}
        r = pack_data(data)
        parts = misc.split_float(x)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.FLOAT_2,
                                  tags.NINT_8, 1,
                                  tags.PINT_32, int(parts.right_significand),
                                  tags.END)
        self.assertEqual(expected, r)

    def test_negative_float_2_ext(self):
        left_significand = 1
        right_significand = 2**32 - 1
        significand = "{}.{}".format(left_significand, right_significand)
        x = float("{}E-10".format(significand))
        data = {0: -x}
        r = pack_data(data)
        parts = misc.split_float(x)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.FLOAT_2_EXT,
                                  tags.NINT_8, 1,
                                  tags.PINT_32, int(parts.right_significand),
                                  tags.NINT_8, abs(int(parts.exponent)),
                                  tags.END)
        self.assertEqual(expected, r)

    def test_positive_float_3(self):
        left_significand = 1
        right_significand = 2 ** 32 - 1
        leading_zeros = "0" * 3
        significand = "{}.{}{}".format(left_significand,
                                       leading_zeros,
                                       right_significand)
        x = float(significand)
        data = {0: x}
        r = pack_data(data)
        parts = misc.split_float(x)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.FLOAT_3,
                                  tags.CONST_1,
                                  tags.CONST_3,
                                  tags.PINT_32, int(parts.right_significand),
                                  tags.END)
        self.assertEqual(expected, r)

    def test_positive_float_3_ext(self):
        left_significand = 1
        right_significand = 2**32 - 1
        leading_zeros = "0" * 3
        significand = "{}.{}{}".format(left_significand,
                                       leading_zeros,
                                       right_significand)
        x = float("{}E-10".format(significand))
        data = {0: x}
        r = pack_data(data)
        parts = misc.split_float(x)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.FLOAT_3_EXT,
                                  tags.CONST_1,
                                  tags.CONST_3,
                                  tags.PINT_32, int(parts.right_significand),
                                  tags.NINT_8, abs(int(parts.exponent)),
                                  tags.END)
        self.assertEqual(expected, r)

    def test_negative_float_3(self):
        left_significand = 1
        right_significand = 2 ** 32 - 1
        leading_zeros = "0" * 3
        significand = "{}.{}{}".format(left_significand,
                                       leading_zeros,
                                       right_significand)
        x = float(significand)
        data = {0: -x}
        r = pack_data(data)
        parts = misc.split_float(x)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.FLOAT_3,
                                  tags.NINT_8, 1,
                                  tags.CONST_3,
                                  tags.PINT_32, int(parts.right_significand),
                                  tags.END)
        self.assertEqual(expected, r)

    def test_negative_float_3_ext(self):
        left_significand = 1
        right_significand = 2**32 - 1
        leading_zeros = "0" * 3
        significand = "{}.{}{}".format(left_significand,
                                       leading_zeros,
                                       right_significand)
        x = float("{}E-10".format(significand))
        data = {0: -x}
        r = pack_data(data)
        parts = misc.split_float(x)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.FLOAT_3_EXT,
                                  tags.NINT_8, 1,
                                  tags.CONST_3,
                                  tags.PINT_32, int(parts.right_significand),
                                  tags.NINT_8, abs(int(parts.exponent)),
                                  tags.END)
        self.assertEqual(expected, r)


class TestComplexNumber(unittest.TestCase):

    def test_with_positive_numbers(self):
        x = 1.2+3.4j
        data = {0: x}
        r = pack_data(data)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.COMPLEX,
                                  tags.FLOAT_2,
                                  tags.CONST_1,
                                  tags.CONST_2,
                                  tags.FLOAT_2,
                                  tags.CONST_3,
                                  tags.CONST_4,
                                  tags.END)
        self.assertEqual(expected, r)

    def test_with_negative_numbers(self):
        x = -1.2-3.4j
        data = {0: x}
        r = pack_data(data)
        expected = misc.forge_bin(tags.DICT,
                                  tags.CONST_0,
                                  tags.COMPLEX,
                                  tags.FLOAT_2,
                                  tags.NINT_8, 1,
                                  tags.CONST_2,
                                  tags.FLOAT_2,
                                  tags.NINT_8, 3,
                                  tags.CONST_4,
                                  tags.END)
        self.assertEqual(expected, r)


def pack_data(data, **kwargs):
    return serializer.pack(data, **kwargs)


if __name__ == '__main__':
    unittest.main()



import unittest
from paradict.queue.txt_queue import TxtQueue


class TestQueue(unittest.TestCase):

    def test_with_complete_data(self):
        s = "first line\nsecond line\n"
        queue = TxtQueue()
        r = put_and_get(queue, s)
        self.assertEqual(s, r)
        self.assertEqual(0, len(queue.buffer))

    def test_with_incomplete_data(self):
        s = "first line\nsecond lin"
        queue = TxtQueue()
        r = put_and_get(queue, s)
        self.assertEqual("first line\n", r)
        self.assertNotEqual(0, len(queue.buffer))


def put_and_get(queue, s):
    queue.enqueue(s)
    buffer = list()
    for s in queue.dequeue():
        buffer.append(s)
    return "".join(buffer)


if __name__ == '__main__':
    unittest.main()



import unittest
from paradict import kv, const


class TestSplitFuncInDataMode(unittest.TestCase):

    def test_with_single_quote_and_compact_string(self):
        key = "'this_is_the_key'"
        val = "this_is_the_val"
        sep = ":"
        d = "{key}{sep} {val}".format(key=key, sep=sep, val=val)
        r = kv.split(d)
        expected = kv.Info(key, val, sep, const.DATA_MODE)
        self.assertEqual(expected, r)

    def test_with_single_quote_and_spaced_string(self):
        key = "'this is the key'"
        val = "this is the val"
        sep = ":"
        d = "{key}{sep} {val}".format(key=key, sep=sep, val=val)
        r = kv.split(d)
        expected = kv.Info(key, val, sep, const.DATA_MODE)
        self.assertEqual(expected, r)

    def test_with_double_quote_and_compact_string(self):
        key = '"this_is_the key"'
        val = "this_is_the_val"
        sep = ":"
        d = '{key}{sep} {val}'.format(key=key, sep=sep, val=val)
        r = kv.split(d)
        expected = kv.Info(key, val, sep, const.DATA_MODE)
        self.assertEqual(expected, r)

    def test_with_double_quote_and_spaced_string(self):
        key = '"this is the key"'
        val = "this is the val"
        sep = ":"
        d = '{key}{sep} {val}'.format(key=key, sep=sep, val=val)
        r = kv.split(d)
        expected = kv.Info(key, val, sep, const.DATA_MODE)
        self.assertEqual(expected, r)

    def test_with_escaped_single_quote(self):
        key = r"'this \' is the key'"
        key_bis = r"'this ' is the key'"
        val = "this is the val"
        sep = ":"
        d = "{key}{sep} {val}".format(key=key, sep=sep, val=val)
        r = kv.split(d)
        expected = kv.Info(key_bis, val, sep, const.DATA_MODE)
        self.assertEqual(expected, r)

    def test_with_escaped_double_quote(self):
        key = r'"this \" is the key"'
        key_bis = r'"this " is the key"'
        val = "this is the val"
        sep = ":"
        d = '{key}{sep} {val}'.format(key=key, sep=sep, val=val)
        r = kv.split(d)
        expected = kv.Info(key_bis, val, sep, const.DATA_MODE)
        self.assertEqual(expected, r)

    def test_with_backslash(self):
        key = r'"this \ is the key"'
        val = "this is the val"
        sep = ":"
        d = '{key}{sep} {val}'.format(key=key, sep=sep, val=val)
        r = kv.split(d)
        expected = kv.Info(key, val, sep, const.DATA_MODE)
        self.assertEqual(expected, r)

    def test_with_double_backslash(self):
        key = r'"this \\ is the key"'
        val = "this is the val"
        sep = ":"
        d = '{key}{sep} {val}'.format(key=key, sep=sep, val=val)
        r = kv.split(d)
        expected = kv.Info(key, val, sep, const.DATA_MODE)
        self.assertEqual(expected, r)


if __name__ == '__main__':
    unittest.main()



from paradict.serializer import encode, pack, dump
from paradict.serializer.encoder import Encoder
from paradict.serializer.packer import Packer
from paradict.deserializer import decode, unpack, load
from paradict.deserializer.decoder import Decoder
from paradict.deserializer.unpacker import Unpacker
from paradict.typeref import TypeRef
from paradict.validator import validate, Validator
from paradict.kv import split as split_kv
from paradict.misc import forge_bin, stringify_bin
from paradict.const import CONFIG_MODE, DATA_MODE


__all__ = ["encode", "decode",
           "pack", "unpack",
           "load", "dump",
           "validate", "split_kv",
           "forge_bin", "stringify_bin",
           "Encoder", "Decoder",
           "Packer", "Unpacker",
           "TypeRef", "Validator", "CONFIG_MODE", "DATA_MODE"]



import unittest
import datetime
from decimal import Decimal
from textwrap import dedent
from paradict import serializer
from paradict import box, const, errors


class TestEmptyData(unittest.TestCase):

    def test(self):
        data = dict()
        r = encode_data(data)
        expected = ""
        self.assertEqual(expected, r)


class TestDictInConfigMode(unittest.TestCase):

    def test_valid_dict(self):
        d = {"a_valid_key_42": {"another_key": 42}}
        r = encode_data(d, mode=const.CONFIG_MODE)
        expected = """\
        a_valid_key_42 = (dict)
            another_key = 42"""
        self.assertEqual(dedent(expected), r)

    def test_invalid_dict(self):  # in CONFIG_MODE, only string is allowed as dict key
        d1 = {42: None}
        d2 = {None: None}
        d3 = {True: None}
        d4 = {4.2: None}
        d5 = {complex(1, 2): None}
        d6 = {datetime.datetime(2020, 12, 31): None}
        for i, d in enumerate((d1, d2, d3, d4, d5, d6)):
            with self.subTest("Test {}".format(i + 1)):
                with self.assertRaises(errors.Error):
                    encode_data(d, mode=const.CONFIG_MODE)

    def test_malformed_dict_key(self):
        d1 = {"key_42=": None}
        d2 = {"1_key 42": None}
        for i, d in enumerate((d1, d2)):
            with self.subTest("Test {}".format(i+1)):
                with self.assertRaises(errors.Error):
                    encode_data(d, mode=const.CONFIG_MODE)


class TestDict(unittest.TestCase):

    def test_empty_dict(self):
        d = {0: dict()}
        r = encode_data(d)
        expected = "0: (dict)"
        self.assertEqual(expected, r)

    def test_dict_with_valid_keys(self):
        # are valid keys: integer (int, box.HexInt, ...),
        # float, complex and string
        d = {0: None,
             "1": None,
             "": None,
             "key": None,
             3.14: None,
             complex(1, 2): None,
             "\n": None,
             "\\n": None}
        r = encode_data(d)
        expected = """\
        0: null
        "1": null
        "": null
        "key": null
        3.14: null
        1+2i: null
        "\\n": null
        "\\\\n": null"""
        self.assertEqual(dedent(expected), r)

    def test_dict_with_valid_values(self):
        d = {0: "hello world",
             1: "multiline\nstring",
             2: True,
             3: False,
             4: complex(1, 2),
             5: 42,
             6: 4.2,
             7: datetime.date(2020, 12, 31)}
        r = encode_data(d)
        expected = """\
        0: "hello world"
        1: (text)
            multiline
            string
            ---
        2: true
        3: false
        4: 1+2i
        5: 42
        6: 4.2
        7: 2020-12-31"""
        self.assertEqual(dedent(expected), r)

    def test_dict_with_invalid_keys(self):
        d1 = {None: None}
        d2 = {True: None}
        d3 = {datetime.datetime(2020, 12, 31): None}
        for i, d in enumerate((d1, d2, d3)):
            with self.subTest("Test {}".format(i + 1)):
                with self.assertRaises(errors.Error):
                    encode_data(d, mode=const.CONFIG_MODE)

    def test_nested_dict(self):
        d = {0: "hello world",
             "1": "multiline\nstring",
             "": True,
             "key": False,
             3.14: complex(1, 2),
             complex(1, 2): 42,
             "\n": 4.2,
             "\\n": datetime.date(2020, 12, 31),
             "nested": {0: "hello world",
                        "1": "multiline\nstring",
                        "": True,
                        "key": False,
                        3.14: complex(1, 2),
                        complex(1, 2): 42,
                        "\n": 4.2,
                        "\\n": datetime.date(2020, 12, 31)}}
        r = encode_data(d)
        expected = """\
                0: "hello world"
                "1": (text)
                    multiline
                    string
                    ---
                "": true
                "key": false
                3.14: 1+2i
                1+2i: 42
                "\\n": 4.2
                "\\\\n": 2020-12-31
                "nested": (dict)
                    0: "hello world"
                    "1": (text)
                        multiline
                        string
                        ---
                    "": true
                    "key": false
                    3.14: 1+2i
                    1+2i: 42
                    "\\n": 4.2
                    "\\\\n": 2020-12-31"""
        self.assertEqual(dedent(expected), r)


class TestList(unittest.TestCase):

    def test_empty_list(self):
        d = {0: list()}
        r = encode_data(d)
        expected = "0: (list)"
        self.assertEqual(expected, r)

    def test_list_with_valid_data(self):
        d = {0: [0, 4.2, "1", True, False, None,
                 complex(1, 2), "hello world",
                 datetime.date(2020, 12, 31),
                 "multiline\ntext"]}
        r = encode_data(d)
        expected = """\
        0: (list)
            0
            4.2
            "1"
            true
            false
            null
            1+2i
            "hello world"
            2020-12-31
            (text)
                multiline
                text
                ---"""
        self.assertEqual(dedent(expected), r)

    def test_nested_list(self):
        d = {0: [0, 4.2, "1", True, False, None,
                 complex(1, 2), "hello world",
                 datetime.date(2020, 12, 31),
                 "multiline\ntext",
                 [0, 4.2, "1", True, False, None,
                  complex(1, 2), "hello world",
                  datetime.date(2020, 12, 31),
                  "multiline\ntext"]]}
        r = encode_data(d)
        expected = """\
        0: (list)
            0
            4.2
            "1"
            true
            false
            null
            1+2i
            "hello world"
            2020-12-31
            (text)
                multiline
                text
                ---
            (list)
                0
                4.2
                "1"
                true
                false
                null
                1+2i
                "hello world"
                2020-12-31
                (text)
                    multiline
                    text
                    ---"""
        self.assertEqual(dedent(expected), r)


class TestSet(unittest.TestCase):

    def test_empty_set(self):
        d = {0: set()}
        r = encode_data(d)
        expected = "0: (set)"
        self.assertEqual(expected, r)

    def test_set_with_valid_data(self):
        d = {0: {0, 4.2, "1",
                 complex(1, 2), "hello world",
                 datetime.date(2020, 12, 31),
                 "multiline\ntext"}}
        r = encode_data(d)
        expected = """\
        0: (set)
            0
            4.2
            "1"
            1+2i
            "hello world"
            2020-12-31
            (text)
                multiline
                text
                ---"""
        self.assertEqual(len(dedent(expected)), len(r))

    def test_set_with_invalid_data(self):
        d = {0: {0, 4.2, "1", True, False, None,
                 complex(1, 2), "hello world",
                 datetime.date(2020, 12, 31),
                 "multiline\ntext"}}
        with self.assertRaises(errors.Error):
            encode_data(d)


class TestObj(unittest.TestCase):

    def test_empty_obj(self):
        d = {0: box.Obj()}
        r = encode_data(d)
        expected = "0: (obj)"
        self.assertEqual(expected, r)

    def test_obj_with_valid_keys(self):
        # are valid keys: integer (int, box.HexInt, ...),
        # float, complex and string (str, box.Raw)
        x = {0: None,
             "1": None,
             "": None,
             "key": None,
             3.14: None,
             complex(1, 2): None,
             "\n": None,
             "\\n": None}
        d = {0: box.Obj(x)}
        r = encode_data(d)
        expected = """\
        0: (obj)
            0: null
            "1": null
            "": null
            "key": null
            3.14: null
            1+2i: null
            "\\n": null
            "\\\\n": null"""
        self.assertEqual(dedent(expected), r)

    def test_obj_with_valid_values(self):
        x = {0: "hello world",
             1: "multiline\nstring",
             2: True,
             3: False,
             4: complex(1, 2),
             5: 42,
             6: 4.2,
             7: datetime.date(2020, 12, 31)}
        d = {0: box.Obj(x)}
        r = encode_data(d)
        expected = """\
        0: (obj)
            0: "hello world"
            1: (text)
                multiline
                string
                ---
            2: true
            3: false
            4: 1+2i
            5: 42
            6: 4.2
            7: 2020-12-31"""
        self.assertEqual(dedent(expected), r)

    def test_obj_with_invalid_keys(self):
        x1 = {None: None}
        x2 = {True: None}
        x3 = {datetime.datetime(2020, 12, 31): None}
        for i, x in enumerate((x1, x2, x3)):
            d = {0: box.Obj(x)}
            with self.subTest("Test {}".format(i + 1)):
                with self.assertRaises(errors.Error):
                    encode_data(d, mode=const.CONFIG_MODE)

    def test_nested_obj(self):
        x = {0: "hello world",
             "1": "multiline\nstring",
             "": True,
             "key": False,
             3.14: complex(1, 2),
             complex(1, 2): 42,
             "\n": 4.2,
             "\\n": datetime.date(2020, 12, 31),
             "nested": box.Obj({0: "hello world",
                        "1": "multiline\nstring",
                        "": True,
                        "key": False,
                        3.14: complex(1, 2),
                        complex(1, 2): 42,
                        "\n": 4.2,
                        "\\n": datetime.date(2020, 12, 31)})}
        d = {0: box.Obj(x)}
        r = encode_data(d)
        expected = """\
                0: (obj)
                    0: "hello world"
                    "1": (text)
                        multiline
                        string
                        ---
                    "": true
                    "key": false
                    3.14: 1+2i
                    1+2i: 42
                    "\\n": 4.2
                    "\\\\n": 2020-12-31
                    "nested": (obj)
                        0: "hello world"
                        "1": (text)
                            multiline
                            string
                            ---
                        "": true
                        "key": false
                        3.14: 1+2i
                        1+2i: 42
                        "\\n": 4.2
                        "\\\\n": 2020-12-31"""
        self.assertEqual(dedent(expected), r)


class TestGrid(unittest.TestCase):

    def test_valid_grid(self):
        x = [(0, 1, 2.3, 4+5j, box.HexInt("0xffff_ffff")),
             (0, 1, 2.3, 4+5j, box.BinInt("0b0000_1111")),
             (0, 1, 2.3, 4+5j, box.OctInt("0o0_000_777"))]
        d = {0: box.Grid(x)}
        r = encode_data(d)
        expected = """\
        0: (grid)
            0 1 2.3 4+5i 0xFFFF_FFFF
            0 1 2.3 4+5i 0b0000_1111
            0 1 2.3 4+5i 0o0_000_777"""
        self.assertEqual(dedent(expected), r)

    def test_inconsistent_grid(self):
        x = [(0, 1, 2.3, 4 + 5j, box.HexInt("0xff"), box.HexInt("0xff")),
             (0, 1, 2.3, 4 + 5j, box.HexInt("0xff")),
             (0, 1, 2.3, 4 + 5j, box.HexInt("0xff"))]
        d = {0: box.Grid(x)}
        with self.assertRaises(errors.Error):
            encode_data(d)

    def test_invalid_grid(self):
        x = [("a", "b", "c", "d", box.HexInt("0xff")),
             (0, 1, 2.3, 4 + 5j, box.HexInt("0xff")),
             (0, 1, 2.3, 4 + 5j, box.HexInt("0xff"))]
        d = {0: box.Grid(x)}
        with self.assertRaises(errors.Error):
            encode_data(d)


class TestBool(unittest.TestCase):

    def test_true(self):
        d = {0: True}
        r = encode_data(d)
        expected = "0: true"
        self.assertEqual(expected, r)

    def test_false(self):
        d = {0: False}
        r = encode_data(d)
        expected = "0: false"
        self.assertEqual(expected, r)


class TestNull(unittest.TestCase):

    def test(self):
        d = {0: None}
        r = encode_data(d)
        expected = "0: null"
        self.assertEqual(expected, r)


class TestDatetime(unittest.TestCase):

    def test_naive_datetime_lower_endpoint(self):
        year, month = 1, 1
        day, hour, minute = 1, 0, 0
        second, microsecond = 0, 0
        val = datetime.datetime(year=year, month=month, day=day,
                                hour=hour, minute=minute,
                                second=second, microsecond=microsecond)
        d = {0: val}
        r = encode_data(d)
        expected = "0: 0001-01-01T00:00:00"
        self.assertEqual(expected, r)

    def test_naive_datetime_upper_endpoint(self):
        year, month = 9999, 12
        day, hour, minute = 31, 23, 59
        second, microsecond = 59, 999999
        val = datetime.datetime(year=year, month=month, day=day,
                                hour=hour, minute=minute,
                                second=second, microsecond=microsecond)
        d = {0: val}
        r = encode_data(d)
        expected = "0: 9999-12-31T23:59:59.999999"
        self.assertEqual(expected, r)

    def test_aware_datetime_lower_endpoint(self):
        year, month = 1, 1
        day, hour, minute = 1, 0, 0
        second, microsecond = 0, 0
        tz = datetime.timezone(datetime.timedelta(hours=0, minutes=0))
        val = datetime.datetime(year=year, month=month, day=day,
                                hour=hour, minute=minute,
                                second=second, microsecond=microsecond,
                                tzinfo=tz)
        d = {0: val}
        r = encode_data(d)
        expected = "0: 0001-01-01T00:00:00Z"
        self.assertEqual(expected, r)

    def test_aware_datetime_upper_endpoint(self):
        year, month = 9999, 12
        day, hour, minute = 31, 23, 59
        second, microsecond = 59, 999999
        tz = datetime.timezone(datetime.timedelta(hours=23, minutes=59))
        val = datetime.datetime(year=year, month=month, day=day,
                                hour=hour, minute=minute,
                                second=second, microsecond=microsecond,
                                tzinfo=tz)
        d = {0: val}
        r = encode_data(d)
        expected = "0: 9999-12-31T23:59:59.999999+23:59"
        self.assertEqual(expected, r)

    def test_aware_datetime_with_negative_negative_utc_timezone(self):
        year, month = 9999, 12
        day, hour, minute = 31, 23, 59
        second, microsecond = 59, 999999
        tz = datetime.timezone(datetime.timedelta(hours=-1, minutes=-30))
        val = datetime.datetime(year=year, month=month, day=day,
                                hour=hour, minute=minute,
                                second=second, microsecond=microsecond,
                                tzinfo=tz)
        d = {0: val}
        r = encode_data(d)
        expected = "0: 9999-12-31T23:59:59.999999-01:30"
        self.assertEqual(expected, r)


class TestDate(unittest.TestCase):

    def test_date_lower_endpoint(self):
        year, month, day = 1, 1, 1
        val = datetime.date(year=year, month=month, day=day)
        d = {0: val}
        r = encode_data(d)
        expected = "0: 0001-01-01"
        self.assertEqual(expected, r)

    def test_date_upper_endpoint(self):
        year, month, day = 9999, 12, 31
        val = datetime.date(year=year, month=month, day=day)
        d = {0: val}
        r = encode_data(d)
        expected = "0: 9999-12-31"
        self.assertEqual(expected, r)


class TestTime(unittest.TestCase):

    def test_naive_time_lower_endpoint(self):
        hour, minute = 0, 0
        second, microsecond = 0, 0
        val = datetime.time(hour=hour, minute=minute,
                            second=second, microsecond=microsecond)
        d = {0: val}
        r = encode_data(d)
        expected = "0: 00:00:00"
        self.assertEqual(expected, r)

    def test_naive_time_upper_endpoint(self):
        hour, minute = 23, 59
        second, microsecond = 59, 999999
        val = datetime.time(hour=hour, minute=minute,
                            second=second, microsecond=microsecond)
        d = {0: val}
        r = encode_data(d)
        expected = "0: 23:59:59.999999"
        self.assertEqual(expected, r)

    def test_aware_time_lower_endpoint(self):
        hour, minute = 0, 0
        second, microsecond = 0, 0
        tz = datetime.timezone(datetime.timedelta(hours=0, minutes=0))
        val = datetime.time(hour=hour, minute=minute,
                            second=second, microsecond=microsecond,
                            tzinfo=tz)
        d = {0: val}
        r = encode_data(d)
        expected = "0: 00:00:00+00:00"
        self.assertEqual(expected, r)

    def test_aware_time_upper_endpoint(self):
        hour, minute = 23, 59
        second, microsecond = 59, 999999
        tz = datetime.timezone(datetime.timedelta(hours=23, minutes=59))
        val = datetime.time(hour=hour, minute=minute,
                            second=second, microsecond=microsecond,
                            tzinfo=tz)
        d = {0: val}
        r = encode_data(d)
        expected = "0: 23:59:59.999999+23:59"
        self.assertEqual(expected, r)

    def test_aware_time_with_negative_negative_utc_timezone(self):
        hour, minute = 23, 59
        second, microsecond = 59, 999999
        tz = datetime.timezone(datetime.timedelta(hours=-1, minutes=-30))
        val = datetime.time(hour=hour, minute=minute,
                            second=second, microsecond=microsecond,
                            tzinfo=tz)
        d = {0: val}
        r = encode_data(d)
        expected = "0: 23:59:59.999999-01:30"
        self.assertEqual(expected, r)


class TestInteger(unittest.TestCase):

    def test_positive_int(self):
        d = {0: 10_000_000_000}
        r = encode_data(d)
        expected = "0: 10_000_000_000"
        self.assertEqual(expected, r)

    def test_negative_int(self):
        d = {0: -10_000_000_000}
        r = encode_data(d)
        expected = "0: -10_000_000_000"
        self.assertEqual(expected, r)

    def test_positive_hex_int(self):
        x = 2**32-1
        d = {0: box.HexInt(x)}
        r = encode_data(d)
        expected = "0: 0xFFFF_FFFF"
        self.assertEqual(expected, r)

    def test_negative_hex_int(self):
        x = -(2**32-1)
        d = {0: box.HexInt(x)}
        r = encode_data(d)
        expected = "0: -0xFFFF_FFFF"
        self.assertEqual(expected, r)

    def test_positive_oct_int(self):
        x = 2**32-1
        d = {0: box.OctInt(x)}
        r = encode_data(d)
        expected = "0: 0o37_777_777_777"
        self.assertEqual(expected, r)

    def test_negative_oct_int(self):
        x = -(2**32-1)
        d = {0: box.OctInt(x)}
        r = encode_data(d)
        expected = "0: -0o37_777_777_777"
        self.assertEqual(expected, r)

    def test_positive_bin_int(self):
        x = 2**8-1
        d = {0: box.BinInt(x)}
        r = encode_data(d)
        expected = "0: 0b1111_1111"
        self.assertEqual(expected, r)

    def test_negative_bin_int(self):
        x = -(2**8-1)
        d = {0: box.BinInt(x)}
        r = encode_data(d)
        expected = "0: -0b1111_1111"
        self.assertEqual(expected, r)


class TestIntWithLeadingZeros(unittest.TestCase):

    def test_hex(self):
        d = {0: box.HexInt("-0x0000_FFFF")}
        r = encode_data(d)
        expected = "0: -0x0000_FFFF"
        self.assertEqual(expected, r)

    def test_oct(self):
        d = {0: box.OctInt("-0o000_777")}
        r = encode_data(d)
        expected = "0: -0o000_777"
        self.assertEqual(expected, r)

    def test_bin(self):
        d = {0: box.BinInt("-0b0000_1111")}
        r = encode_data(d)
        expected = "0: -0b0000_1111"
        self.assertEqual(expected, r)


class TestFloat(unittest.TestCase):

    def test_positive_float(self):
        d = {0: 10_000_000_000.000_01}
        r = encode_data(d)
        expected = "0: 10_000_000_000.000_01"
        self.assertEqual(expected, r)

    def test_negative_float(self):
        d = {0: -10_000_000_000.000_01}
        r = encode_data(d)
        expected = "0: -10_000_000_000.000_01"
        self.assertEqual(expected, r)

    def test_positive_float_with_scientific_notation(self):
        d = {0: float("1e-10")}
        r = encode_data(d)
        expected = "0: 1.0E-10"
        self.assertEqual(expected, r)

    def test_negative_float_with_scientific_notation(self):
        d = {0: float("-1e-10")}
        r = encode_data(d)
        expected = "0: -1.0E-10"
        self.assertEqual(expected, r)


class TestDecimalFloat(unittest.TestCase):

    def test_positive_float(self):
        d = {0: Decimal("1.000_234_56")}
        r = encode_data(d)
        expected = "0: 1.000_234_56"
        self.assertEqual(expected, r)

    def test_negative_float(self):
        d = {0: Decimal("-1.000_234_56")}
        r = encode_data(d)
        expected = "0: -1.000_234_56"
        self.assertEqual(expected, r)

    def test_positive_float_with_scientific_notation(self):
        d = {0: Decimal("1.000_234_56E10")}
        r = encode_data(d)
        expected = "0: 1.000_234_56E10"
        self.assertEqual(expected, r)

    def test_negative_float_with_scientific_notation(self):
        d = {0: Decimal("-1.000_234_56E-10")}
        r = encode_data(d)
        expected = "0: -1.000_234_56E-10"
        self.assertEqual(expected, r)


class TestMultilineNumber(unittest.TestCase):

    def test_int(self):
        d = {0: -1_000_000_000_000_000_000_000_000_000_000_000_000_000_000_000_000_000_000}
        r = encode_data(d)
        expected = """\
        0: (int)
            -1_000_000_000_000_000_000_000_000_000_000
            _000_000_000_000_000_000_000_000"""
        self.assertEqual(dedent(expected), r)

    def test_hex_int(self):
        d = {0: box.HexInt("-0x100_0000_0000_0000_0000_0000_0000_0000_0000_0000_0000_0000_0000_0000")}
        r = encode_data(d)
        expected = """\
        0: (int)
            -0x100_0000_0000_0000_0000_0000_0000_0000_
            0000_0000_0000_0000_0000_0000"""
        self.assertEqual(dedent(expected), r)

    def test_oct_int(self):
        d = {0: box.OctInt("-0o1_000_000_000_000_000_000_000_000_000_000_000_000_000_000_000_000_000_000")}
        r = encode_data(d)
        expected = """\
        0: (int)
            -0o1_000_000_000_000_000_000_000_000_000_0
            00_000_000_000_000_000_000_000_000"""
        self.assertEqual(dedent(expected), r)

    def test_bin_int(self):
        d = {0: box.BinInt("-0b100_0000_0000_0000_0000_0000_0000_0000_0000_0000_0000_0000_0000_0000")}
        r = encode_data(d)
        expected = """\
        0: (int)
            -0b100_0000_0000_0000_0000_0000_0000_0000_
            0000_0000_0000_0000_0000_0000"""
        self.assertEqual(dedent(expected), r)


    def test_float(self):
        d = {0: Decimal("-1.000_000_000_000_000_000_000_000_000_000_000_000_000_000_000_000_000_111E-10")}
        r = encode_data(d)
        expected = """\
        0: (float)
            -1.000_000_000_000_000_000_000_000_000_000
            _000_000_000_000_000_000_000_111E-10"""
        self.assertEqual(dedent(expected), r)


class TestComplexNumber(unittest.TestCase):

    def test_1(self):
        d = {0: complex(0, 1)}
        r = encode_data(d)
        expected = "0: 0+1i"
        self.assertEqual(expected, r)

    def test_2(self):
        d = {0: complex(-1, -1)}
        r = encode_data(d)
        expected = "0: -1-1i"
        self.assertEqual(expected, r)

    def test_3(self):
        d = {0: complex(-1.0, -1000.0)}
        r = encode_data(d)
        expected = "0: -1-1_000i"
        self.assertEqual(expected, r)

    def test_4(self):
        d = {0: complex(-1.234_567_89, -1000.0)}
        r = encode_data(d)
        expected = "0: -1.234_567_89-1_000i"
        self.assertEqual(expected, r)

    def test_5(self):
        d = {0: complex(-1.234_567_89E-10, 1.234_567_89E-10)}
        r = encode_data(d)
        expected = "0: -1.234_567_89E-10+1.234_567_89E-10i"
        self.assertEqual(expected, r)


class TestString(unittest.TestCase):

    def test_empty_str(self):
        d = {0: ""}
        r = encode_data(d)
        expected = '0: ""'
        self.assertEqual(expected, r)

    def test_str(self):
        d = {0: "hello \u02eb world"}
        r = encode_data(d)
        expected = '0: "hello ˫ world"'
        self.assertEqual(expected, r)

    def test_str_with_backlash(self):
        d = {0: "hello \u02eb \\ world"}
        r = encode_data(d)
        expected = r"0: 'hello ˫ \ world'"
        self.assertEqual(expected, r)


class TestMultilineString(unittest.TestCase):

    def test_str(self):
        d = {0: "this is \u02eb a\nmultiline string"}
        r = encode_data(d)
        expected = """\
        0: (text)
            this is ˫ a
            multiline string
            ---"""
        self.assertEqual(dedent(expected), r)

    def test_str_with_backlash(self):
        d = {0: "this is \u02eb a\nmultiline \\string"}
        r = encode_data(d)
        expected = """\
        0: (raw)
            this is ˫ a
            multiline \\string
            ---"""
        self.assertEqual(dedent(expected), r)


class TestComment(unittest.TestCase):

    def test_empty_comment(self):
        c = ""
        d = {box.CommentID(): box.Comment(c),
             box.CommentID(): box.Comment(c),
             0: {box.CommentID(): box.Comment(c),
                 box.CommentID(): box.Comment(c),
                 0: [box.Comment(c), box.Comment(c),
                     box.Obj({box.CommentID(): box.Comment(c),
                              box.CommentID(): box.Comment(c),
                              0: {box.Comment(c)}})]}}
        r = encode_data(d, skip_comments=False)
        expected = """\
        #
        #
        0: (dict)
            #
            #
            0: (list)
                #
                #
                (obj)
                    #
                    #
                    0: (set)
                        #"""
        self.assertEqual(dedent(expected), r)

    def test_regular_comment(self):
        c1 = "comment 1"
        c2 = "comment 2"
        c3 = "comment 3"
        c4 = "comment 4"
        c5 = "comment 5"
        d = {box.CommentID(): box.Comment(c1),
             0: {box.CommentID(): box.Comment(c2),
                 0: [box.Comment(c3),
                     box.Obj({box.CommentID(): box.Comment(c4),
                              0: {box.Comment(c5)}})]}}
        r = encode_data(d, skip_comments=False)
        expected = """\
        # comment 1
        0: (dict)
            # comment 2
            0: (list)
                # comment 3
                (obj)
                    # comment 4
                    0: (set)
                        # comment 5"""
        self.assertEqual(dedent(expected), r)


class TestBin(unittest.TestCase):

    def test_empty_bin_data(self):
        d = {0: b""}
        r = encode_data(d, skip_bin_data=False)
        expected = """0: (bin)"""
        self.assertEqual(dedent(expected), r)

    def test_bin_data(self):
        d = {0: b"hello world hello world hello world hello world hello world hello world hello world"}
        r = encode_data(d, skip_bin_data=False)
        expected = """\
        0: (bin)
            68 65 6C 6C 6F 20 77 6F 72 6C 64 20 68 65 6C 6C
            6F 20 77 6F 72 6C 64 20 68 65 6C 6C 6F 20 77 6F
            72 6C 64 20 68 65 6C 6C 6F 20 77 6F 72 6C 64 20
            68 65 6C 6C 6F 20 77 6F 72 6C 64 20 68 65 6C 6C
            6F 20 77 6F 72 6C 64 20 68 65 6C 6C 6F 20 77 6F
            72 6C 64"""
        self.assertEqual(dedent(expected), r)

    def test_skipped_bin_data(self):
        d = {0: b"hello world hello world hello world hello world hello world hello world hello world"}
        r = encode_data(d, skip_bin_data=True)
        expected = """\
        0: (bin)"""
        self.assertEqual(dedent(expected), r)


def encode_data(data, **kwargs):
    return serializer.encode(data, **kwargs)


if __name__ == '__main__':
    unittest.main()



import unittest
from paradict.serializer.packer import Packer
from paradict.deserializer.unpacker import Unpacker


class TestBinStreaming(unittest.TestCase):

    def test(self):
        # This stream is made of messages
        # Each message is a dictionary that serves as envelope
        stream = [{0: "a"}, {0: "b"}, {0: "c"}]
        # Result will hold the unpacked messages
        result = list()
        # instantiate packer and unpacker
        packer = Packer()
        # the receiver takes as argument the reference to the unpacker
        unpacker = Unpacker(receiver=lambda ref: result.append(ref.data))
        # iterate over the stream to pack each message into datums
        # that will feed the unpacker which will call the receiver
        # after each complete unpacking of a message.
        # The unpacker holds a reference to the latest
        # unpacked message via the "unpacker.data" property
        for i, msg in enumerate(stream):
            for datum in packer.pack(msg):
                unpacker.feed(datum)
            # check if datum is well unpacked
            with self.subTest("Message At Index: {}".format(i)):
                # unpacker.data holds unpacked data
                self.assertEqual(msg, unpacker.data)
        # check if the original stream contents is mirrored in
        # the result variable
        self.assertEqual(stream, result)


if __name__ == '__main__':
    unittest.main()



"""Private miscellaneous module for the Paradict binary format"""
from paradict import tags


# Mapping bytes number to INT tags. Eg: 1 byte int belongs to tags.INT_8
SIZE_TO_PINT = {1: tags.PINT_8, 2: tags.PINT_16,
                3: tags.PINT_24, 4: tags.PINT_32,
                5: tags.PINT_40, 6: tags.PINT_48,
                7: tags.PINT_56, 8: tags.PINT_64}


# Mapping bytes number to NINT tags. Eg: 1 byte nint belongs to tags.NINT_8
SIZE_TO_NINT = {1: tags.NINT_8, 2: tags.NINT_16,
                3: tags.NINT_24, 4: tags.NINT_32,
                5: tags.NINT_40, 6: tags.NINT_48,
                7: tags.NINT_56, 8: tags.NINT_64}


# Mapping bytes number to string tags. Eg: 1 byte string belongs to tags.STR_8
SIZE_TO_STR = {1: tags.STR_8, 2: tags.STR_16,
               3: tags.STR_24, 4: tags.STR_32,
               5: tags.STR_40, 6: tags.STR_48,
               7: tags.STR_56, 8: tags.STR_64,
               9: tags.STR_72, 10: tags.STR_80,
               11: tags.STR_88, 12: tags.STR_96,
               13: tags.STR_104, 14: tags.STR_112,
               15: tags.STR_120, 16: tags.STR_128,
               17: tags.STR_136, 18: tags.STR_144,
               19: tags.STR_152, 20: tags.STR_160,
               21: tags.STR_168, 22: tags.STR_176,
               23: tags.STR_184, 24: tags.STR_192,
               25: tags.STR_200, 26: tags.STR_208,
               27: tags.STR_216, 28: tags.STR_224,
               29: tags.STR_232, 30: tags.STR_240,
               31: tags.STR_248, 32: tags.STR_256}


# Mapping lower and upper alphabet letters to their tags. Eg: "a" -> tags.LOWER_A
ALPHABET = {"a": tags.CHAR_A, "b": tags.CHAR_B,
            "c": tags.CHAR_C, "d": tags.CHAR_D,
            "e": tags.CHAR_E, "f": tags.CHAR_F,
            "g": tags.CHAR_G, "h": tags.CHAR_H,
            "i": tags.CHAR_I, "j": tags.CHAR_J,
            "k": tags.CHAR_K, "l": tags.CHAR_L,
            "m": tags.CHAR_M, "n": tags.CHAR_N,
            "o": tags.CHAR_O, "p": tags.CHAR_P,
            "q": tags.CHAR_Q, "r": tags.CHAR_R,
            "s": tags.CHAR_S, "t": tags.CHAR_T,
            "u": tags.CHAR_U, "v": tags.CHAR_V,
            "w": tags.CHAR_W, "x": tags.CHAR_X,
            "y": tags.CHAR_Y, "z": tags.CHAR_Z,
            "A": tags.CHAR_UP_A, "B": tags.CHAR_UP_B,
            "C": tags.CHAR_UP_C, "D": tags.CHAR_UP_D,
            "E": tags.CHAR_UP_E, "F": tags.CHAR_UP_F,
            "G": tags.CHAR_UP_G, "H": tags.CHAR_UP_H,
            "I": tags.CHAR_UP_I, "J": tags.CHAR_UP_J,
            "K": tags.CHAR_UP_K, "L": tags.CHAR_UP_L,
            "M": tags.CHAR_UP_M, "N": tags.CHAR_UP_N,
            "O": tags.CHAR_UP_O, "P": tags.CHAR_UP_P,
            "Q": tags.CHAR_UP_Q, "R": tags.CHAR_UP_R,
            "S": tags.CHAR_UP_S, "T": tags.CHAR_UP_T,
            "U": tags.CHAR_UP_U, "V": tags.CHAR_UP_V,
            "W": tags.CHAR_UP_W, "X": tags.CHAR_UP_X,
            "Y": tags.CHAR_UP_Y, "Z": tags.CHAR_UP_Z}
